#include <fcntl.h>
#include <linux/videodev.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include "palettes.h"
#include "videolib.h"


// thanks ned!
#define EINTR_RETRY(expression)                                                   \
   (__extension__                                                                 \
      ({                                                                          \
         long int __result;                                                       \
         do {                                                                     \
            sigset_t ss1, ss2;                                                    \
                                                                                  \
            sigemptyset(&ss1);          /* init empty set of signals */           \
            sigaddset(&ss1, SIGALRM);   /* add ALRM to set of blocked signals */  \
            /* can add more signals to block here */                              \
            sigprocmask(SIG_BLOCK, &ss1, &ss2); /* block some signals */          \
                                                                                  \
            __result = (long int) (expression);                                   \
                                                                                  \
            sigprocmask(SIG_SETMASK, &ss2, NULL); /* restore */                   \
         }                                                                        \
         while ((__result == -1L) && (errno == EINTR));                           \
         __result;                                                                \
      })                                                                          \
   )


void showCameraInformation(Camera camera) {
   fprintf(stderr, "    - Camera Information:\n");
   fprintf(stderr, "      ================================================================\n");
   fprintf(stderr, "       Name: %s\n", camera->vcapability.name);
   fprintf(stderr, "       Type: %d\n", camera->vcapability.type);

   if ((camera->vcapability.type & VID_TYPE_CAPTURE) != 0) {
      fprintf(stderr, "             Can capture to memory\n");
   }
   if ((camera->vcapability.type & VID_TYPE_TUNER) != 0) {
      fprintf(stderr, "             Has a tuner of some form\n");
   }
   if ((camera->vcapability.type & VID_TYPE_TELETEXT) != 0) {
      fprintf(stderr, "             Has teletext capability\n");
   }
   if ((camera->vcapability.type & VID_TYPE_OVERLAY) != 0) {
      fprintf(stderr, "             Can overlay its image onto the frame buffer\n");
   }
   if ((camera->vcapability.type & VID_TYPE_CHROMAKEY) != 0) {
      fprintf(stderr, "             Overlay is Chromakeyed\n");
   }
   if ((camera->vcapability.type & VID_TYPE_CLIPPING) != 0) {
      fprintf(stderr, "             Overlay clipping is supported\n");
   }
   if ((camera->vcapability.type & VID_TYPE_FRAMERAM) != 0) {
      fprintf(stderr, "             Overlay overwrites frame buffer memory\n");
   }
   if ((camera->vcapability.type & VID_TYPE_SCALES) != 0) {
      fprintf(stderr, "             The hardware supports image scaling\n");
   }
   if ((camera->vcapability.type & VID_TYPE_MONOCHROME) != 0) {
      fprintf(stderr, "             Image capture is grey scale only\n");
   }
   if ((camera->vcapability.type & VID_TYPE_SUBCAPTURE) != 0) {
      fprintf(stderr, "             Capture can be of only part of the image\n");
   }
   fprintf(stderr,
           "       Channels: %d  Audios: %d  MinExtent: %d@%d  MaxExtent: %d@%d\n",
           camera->vcapability.channels,
           camera->vcapability.audios,
           camera->vcapability.minwidth,
           camera->vcapability.minheight,
           camera->vcapability.maxwidth,
           camera->vcapability.maxheight);
   fprintf(stderr, "      ================================================================\n");

}


void initCamera(Camera camera,
                char* deviceName,
                int width, int height,
                int depth, int palette) {
   camera->deviceName = malloc(1024);
   memset(camera->deviceName, 0, 1024);
   strcpy(camera->deviceName, deviceName);

   camera->desiredWidth  = width;
   camera->desiredHeight = height;

   camera->desiredDepth   = depth;
   camera->desiredPalette = palette;

   camera->fd = 0;

   /*
     camera->vcapability;
     camera->vwindow;
     camera->vpicture;
   */

   camera->buffer = 0;

   /*
     camera->memoryBuffer;
   */
   camera->memoryMap   = 0;
   camera->mmaps       = 0;
   camera->imageSize   = 0;
   camera->bufferIndex = 0;
   camera->buffer24    = 0;

   camera->forceRead = FALSE;
   camera->usingMMap = FALSE;
}


void setupMMap(Camera camera) {
   int i;
   int mmapsSize;
   camera->usingMMap = FALSE;

   fprintf(stderr, "    - Seting up MMAP for camera %p\n", camera);

   if (EINTR_RETRY(ioctl(camera->fd, VIDIOCGMBUF, &camera->memoryBuffer)) < 0) {
      /* failed to retrieve information about capture memory space */
      perror("VIDIOCGMBUF");
      return;
   }

   mmapsSize = camera->memoryBuffer.frames * 32 * 4 /* sizeof(struct video_mmap)*/;

   /* temporary (for ever) workaround */
   /*camera->memoryBuffer.frames = 1; */
   fprintf(stderr, "        - The device has %d buffers\n", camera->memoryBuffer.frames);

   /* obtain memory mapped area */
   camera->memoryMap = (char*) mmap(0,
                                    camera->memoryBuffer.size,
                                    PROT_READ /*| PROT_WRITE*/,
                                    MAP_SHARED, camera->fd, 0);
   if (camera->memoryMap < (char*)0) {
      fprintf(stderr, "        * Failed to retrieve pointer to memory mapped area\n");
      return;
   }

   /* allocate structures */
   fprintf(stderr, "        - Allocing %d bytes\n", mmapsSize);
   camera->mmaps = (struct video_mmap*) malloc(mmapsSize);

   if (camera->mmaps == 0) {
      fprintf(stderr, "        * Not enough memory\n");
      return;
   }

   /* fill out the fields */
   for (i = 0; i < camera->memoryBuffer.frames; i++) {
      fprintf(stderr, "        - Creating the mmap #%d\n", i);

      camera->mmaps[i].frame  = i;
      camera->mmaps[i].width  = camera->vwindow.width;
      camera->mmaps[i].height = camera->vwindow.height;
      camera->mmaps[i].format = camera->vpicture.palette;
   }

   camera->usingMMap = TRUE;
   /*camera->usingMMap = FALSE; */
}


BOOLEAN startCaptureInBuffers(Camera camera) {
   int i;
   for (i = 0; i < (camera->memoryBuffer.frames - 1); i++) {
      fprintf(stderr, "- capturing buffer #%d (1st pass)\n", i);
      if (EINTR_RETRY(ioctl(camera->fd, VIDIOCMCAPTURE, &camera->mmaps[i])) < 0) {
         perror("VIDIOCMCAPTURE1");
         return FALSE;
      }
   }

   camera->bufferIndex = camera->memoryBuffer.frames - 1;
   return TRUE;
}


BOOLEAN switchToRead(Camera camera) {
   camera->usingMMap = FALSE;
   camera->buffer    = malloc(camera->imageSize);
   if (!camera->buffer) {
      fprintf(stderr, "* Out of memory.\n");
      return FALSE;
   }

   return TRUE;
}


BOOLEAN setupCamera(Camera camera) {
   BOOLEAN change;
   fprintf(stderr, "- Seting up camera %p\n", camera);


   /* ---------------------------------------------------------------------- */
   /* let's try to change the capture width & height */
   change = FALSE;
   if (EINTR_RETRY(ioctl(camera->fd, VIDIOCGWIN, &camera->vwindow)) < 0) {
      perror("VIDIOCGWIN1");
      return FALSE;
   }

   if (camera->desiredWidth && (camera->vwindow.width != camera->desiredWidth)) {
     fprintf(stderr, "    - Changing the width from %d to %d\n", camera->vwindow.width, camera->desiredWidth);

     camera->vwindow.width = camera->desiredWidth;
     change = TRUE;
   }
   if (camera->desiredHeight && (camera->vwindow.height != camera->desiredHeight)) {
     fprintf(stderr, "    - Changing the height from %d to %d\n", camera->vwindow.height, camera->desiredHeight);

     camera->vwindow.height = camera->desiredHeight;
     change = TRUE;
   }

   if (change) {
      if (EINTR_RETRY(ioctl(camera->fd, VIDIOCSWIN, &camera->vwindow)) < 0) {
         perror("VIDIOCGWIN2");
         return FALSE;
      }

      if (EINTR_RETRY(ioctl(camera->fd, VIDIOCGWIN, &camera->vwindow)) < 0) {
         perror("VIDIOCGWIN3");
         return FALSE;
      }
      if (camera->desiredWidth != camera->vwindow.width) {
         fprintf(stderr,
                 "    * The device can't change the capture width (now=%d)\n",
                 camera->vwindow.width);
         return FALSE;
      }
      if (camera->desiredHeight != camera->vwindow.height) {
         fprintf(stderr,
                 "    * The device can't change the capture height (now=%d)\n",
                 camera->vwindow.height);
         return FALSE;
      }

   }

   if (EINTR_RETRY(ioctl(camera->fd, VIDIOCGPICT, &camera->vpicture)) < 0) {
      perror("VIDIOCGPICT1");
      return FALSE;
   }

   fprintf(stderr,
           "    - Extent=%d@%d\n",
           camera->vwindow.width,
           camera->vwindow.height);
   /* ---------------------------------------------------------------------- */


   /* ---------------------------------------------------------------------- */
   /* let's try to change the bit depth and palette */
   change = FALSE;
   if (camera->desiredDepth && (camera->vpicture.depth != camera->desiredDepth)) {
     fprintf(stderr, "    - Changing the depth from %d to %d\n", camera->vpicture.depth, camera->desiredDepth);

     camera->vpicture.depth = camera->desiredDepth;
     change = TRUE;
   }
   if (camera->desiredPalette && (camera->vpicture.palette != camera->desiredPalette)) {
     fprintf(stderr,
             "    - Changing the palette from %d (%s) to %d (%s)\n",
             camera->vpicture.palette,
             paletteName(camera->vpicture.palette),
             camera->desiredPalette,
             paletteName(camera->desiredPalette));

     camera->vpicture.palette = camera->desiredPalette;
     change = TRUE;
   }

   if (change) {
      if (EINTR_RETRY(ioctl(camera->fd, VIDIOCSPICT, &camera->vpicture)) < 0) {
         perror("VIDIOCSPICT1");
         return FALSE;
      }

      if (EINTR_RETRY(ioctl(camera->fd, VIDIOCGPICT, &camera->vpicture)) < 0) {
         perror("VIDIOCSPICT2");
         return FALSE;
      }
      if (camera->desiredDepth != camera->vpicture.depth) {
         fprintf(stderr, "    * The device can't change the depth\n");
         return FALSE;
      }
      if (camera->desiredPalette != camera->vpicture.palette) {
         fprintf(stderr, "    * The device can't change the palette\n");
         return FALSE;
      }
   }

   fprintf(stderr,
           "    - Depth=%d Palette=%d %s\n",
           camera->vpicture.depth,
           camera->vpicture.palette,
           paletteName(camera->vpicture.palette));
   /* ---------------------------------------------------------------------- */

   camera->imageSize = camera->vwindow.width * camera->vwindow.height * (paletteBytesPerPixel(camera->vpicture.palette));
   camera->buffer24  = malloc(camera->vwindow.width * camera->vwindow.height * 3);

   if (camera->forceRead) {
      fprintf(stderr, "    - Capturing using (forced) read()\n");
      if (!switchToRead(camera)) {
	 return FALSE;
      }
   }
   else {
      setupMMap(camera);

      if (camera->usingMMap) {
         fprintf(stderr, "    - Capturing using mmap()\n");
         if (!startCaptureInBuffers(camera)) {
            fprintf(stderr, "    * Falling back to read()\n");
	    if (!switchToRead(camera)) {
	       return FALSE;
	    }
         }
      }
      else {
         fprintf(stderr, "    - Capturing using read()\n");
	 if (!switchToRead(camera)) {
	    return FALSE;
	 }
      }
   }

   return TRUE;
}


BOOLEAN openCamera(Camera camera) {
   fprintf(stderr, "- Opening camera in %p\n", camera);

   /* open the video device */
   camera->fd = EINTR_RETRY(open(camera->deviceName, O_RDONLY /*| O_NONBLOCK */));
   if (camera->fd < 0) {
      perror(camera->deviceName);
      return FALSE;
   }
   fprintf(stderr, "    - Opened %s with fd=%d\n", camera->deviceName, camera->fd);

   /* check for a valid device */
   if (EINTR_RETRY(ioctl(camera->fd, VIDIOCGCAP, &camera->vcapability)) < 0) {
      perror("VIDIOGCAP");
      return FALSE;
   }

   showCameraInformation(camera);

   /* check (again) for a valid device */
   if ((camera->vcapability.type & VID_TYPE_CAPTURE) == 0) {
      fprintf(stderr, "* The device can't capture video\n");
      return FALSE;
   }

   if (!setupCamera(camera)) {
      closeCamera(camera);
      return FALSE;
   }

   return TRUE;
}


Camera createCamera(char* deviceName,
                    int width, int height,
                    int depth, int palette) {
   Camera camera;
   fprintf(stderr,
           "- Creating camera deviceName=%s, extent=%d@%d, depth=%d, palette=%d-%s\n",
           deviceName,
           width, height,
           depth, palette,
           paletteName(palette));

   /* fprintf(stderr, "    - allocing %d bytes for the camera structure\n", SIZE_OF_CAMERA); */
   camera = malloc(SIZE_OF_CAMERA);
   if (camera == 0) {
      fprintf(stderr, "    * Not enough memory\n");
      return 0;
   }
   fprintf(stderr, "    - Structure created in %p\n", camera);

   initCamera(camera,
              deviceName,
              width, height,
              depth, palette);

   if (openCamera(camera)) {
      return camera;
   }
   else {
      return 0;
   }
}


void destroyCamera(Camera camera) {
   fprintf(stderr, "- Destroying camera in %p\n", camera);
   free(camera);
}


BOOLEAN nextFrameMMap(Camera camera) {
   /* fprintf(stderr, "- capturing buffer #%d\n", camera->bufferIndex); */

   /* send a request to begin capturing to the currently indexed buffer */
   if (EINTR_RETRY(ioctl(camera->fd, VIDIOCMCAPTURE, &camera->mmaps[camera->bufferIndex])) < 0) {
      perror("VIDIOCMCAPTURE2");
      return FALSE;
   }

   /* move bufferIndex to the next frame */
   camera->bufferIndex++;
   /* if bufferIndex is indexing beyond the last buffer set it to the first buffer */
   camera->bufferIndex %= camera->memoryBuffer.frames;

   /* fprintf(stderr, "- syncing buffer #%d\n", camera->bufferIndex); */

   /* wait for the currently indexed frame to complete capture */
   if (EINTR_RETRY(ioctl(camera->fd, VIDIOCSYNC, &(camera->mmaps[camera->bufferIndex]))) < 0) {
      /* sync request failed */
      perror("VIDIOCSYNC");
      return FALSE;
   }

   /* fprintf(stderr, "- processing buffer #%d\n", camera->bufferIndex); */
   /* return the address of the frame data for the current buffer index */
   camera->buffer = camera->memoryMap + camera->memoryBuffer.offsets[camera->bufferIndex];

   return TRUE;
}



void closeCamera(Camera camera) {
   fprintf(stderr, "- Closing camera in %p\n", camera);

   if (camera->buffer24) {
      free(camera->buffer24);
      camera->buffer24 = 0;
   }

   if (camera->usingMMap) {
      if (camera->mmaps) {
         free(camera->mmaps);
         camera->mmaps = 0;

         munmap(camera->memoryMap, camera->memoryBuffer.size);
      }
   }
   else {
      if (camera->buffer) {
         free(camera->buffer);
      }
   }

   if (camera->fd) {
      close(camera->fd);
      camera->fd = 0;
   }

   destroyCamera(camera);
}


BOOLEAN nextFrameFromCamera(Camera camera) {
   /* fprintf(stderr, "- Capturing from camera in %p\n", camera); */

   if (camera->usingMMap) {
      if (!nextFrameMMap(camera)) {
         return FALSE;
      }
   }
   else {
      /* fprintf(stderr, "- reading...\n"); */

      if (read(camera->fd, camera->buffer, camera->imageSize) != camera->imageSize) {
         fprintf(stderr, "\nThe device didn't answered the correct bytes.\n");
         return FALSE;
      }
   }

   if (paletteConvert(camera->vpicture.palette,
                      camera->vwindow.width,
                      camera->vwindow.height,
                      camera->buffer,
                      camera->buffer24)) {
      return TRUE;
   }
   else {
      return FALSE;
   }
}


void savePPM(char* fileName, Camera camera) {
   char *src = camera->buffer24;

   char red, green, blue;
   int i;

   FILE* ppmFile = fopen(fileName, "w+");

   fprintf(ppmFile, "P6\n%d %d 255\n", camera->vwindow.width, camera->vwindow.height);

   for (i = 0; i < camera->vwindow.width * camera->vwindow.height; i++) {
      red   = *src++;
      green = *src++;
      blue  = *src++;

      fputc(red,   ppmFile);
      fputc(green, ppmFile);
      fputc(blue,  ppmFile);
   }

   fclose(ppmFile);
}
